import json
import os
import random

import pytest
import sqlalchemy.dialects
import sqlalchemy.exc
from sqlalchemy.orm import Session

import rasax.community.database.utils as db_utils
import rasax.community.sql_migrations as migrations
from rasax.community import utils
from rasax.community.database import Role
from tests.unit import conftest
from tests.unit.conftest import insert_test_user, TEST_USER


# Overwrite fixture to run per function
@pytest.fixture()
def session():
    session_maker = conftest.session_maker_for_tests()
    session = session_maker()

    migrations.run_migrations(session)

    # Insert general test user which can be used in later tests
    insert_test_user(session, TEST_USER)

    return session_maker()


def test_query_result_to_dict():
    query_result = ["zero", "one", "two", "three"]
    fields_query = [("d", True), ("a.c", True), ("a.b", True), ("b.c.d", True)]

    actual = utils.query_result_to_dict(query_result, fields_query)

    assert actual["d"] == query_result[0]
    assert actual["a"]["c"] == query_result[1]
    assert actual["a"]["b"] == query_result[2]
    assert actual["b"]["c"]["d"] == query_result[3]


def test_foreign_keys_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, True)
    with pytest.raises(sqlalchemy.exc.IntegrityError):
        db_connection.execute("DROP table {};".format(Role.__tablename__))


def test_turn_off_foreign_keys_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, False)
    if isinstance(db_connection.dialect, sqlalchemy.dialects.sqlite.dialect):
        db_connection.execute("DROP table {};".format(Role.__tablename__))


def test_turn_off_then_on_foreign_key_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, False)
    db_utils.enforce_sqlite_foreign_key_constraints(raw_connection, True)

    with pytest.raises(sqlalchemy.exc.IntegrityError):
        db_connection.execute("DROP table {};".format(Role.__tablename__))


@pytest.mark.parametrize(
    "before_truncation, expected",
    [(1.0, 1.0), (1.1234, 1.12), (19231231.00, 19231231.00)],
)
def test_truncate_float(before_truncation: float, expected: float):
    assert utils.truncate_float(before_truncation, decimal_places=2) == expected


def test_add_plural_suffix():
    base_string = "found item"
    base_string_with_replacement_field = base_string + "{}"

    # list with single item yields singular case
    assert (
        utils.add_plural_suffix(base_string_with_replacement_field, [1]) == "found item"
    )

    # list with multiple items yields plural case
    assert (
        utils.add_plural_suffix(base_string_with_replacement_field, [1, 2, 3])
        == "found items"
    )


def test_query_result_len():
    count = random.randint(1, 10)
    result = utils.QueryResult([{} for _ in range(count)], count)
    assert len(result) == count


def test_decode_base_64():
    import base64

    expected = "hello"
    encoded = base64.b64encode(expected.encode("utf-8")).decode("utf-8")

    assert utils.decode_base64(encoded) == expected


@pytest.mark.parametrize(
    "original, encoded", [("testing", "dGVzdGluZw=="), ("hello", "aGVsbG8=")]
)
def test_encode_base_64(original, encoded):
    assert utils.encode_base64(original) == encoded


# noinspection PyProtectedMember
def test_db_query_from_environment():
    db_query_env_var = "DB_QUERY"

    # env var not set
    os.environ.pop(db_query_env_var, None)
    assert not db_utils._sql_query_parameters_from_environment()

    # env var correctly set
    query = {"driver": "some driver"}
    os.environ[db_query_env_var] = json.dumps(query)
    assert db_utils._sql_query_parameters_from_environment() == query

    # malformed env var
    malformed_db_query = '{"query": "badly formatted}'
    os.environ[db_query_env_var] = malformed_db_query
    assert not db_utils._sql_query_parameters_from_environment()
